/*
 * FarrarSWAligner.h
 *
 *  Created on: Apr 6, 2010
 *      Author: Oleksiy Kohany <kohany@gmail.com>
 */

#ifndef FARRARALIGNER_H_
#define FARRARALIGNER_H_

#include <emmintrin.h>

#include "ScoreMatrix.hpp"
#include "SWAligner.h"

#define SHORT_BIAS 32768

namespace bio {

   /*
    *
    */
   class FarrarSWAligner: public SWAligner {
      class SwFarrarData: public SwData {
      public:
         __m128i *pvsQueryProf;
         __m128i *pvH1;
         __m128i *pvH2;
         __m128i *pvE;
         __m128i *pvEGLLoad;
         __m128i *pvEGLStore;
         __m128i *pvDIR;
         SwFarrarData() {
         }
         virtual ~SwFarrarData() {
         }
         void alloc(size_t queryLength, size_t alphaSize) {
            size_t lenQryShort = (queryLength + 7) / 8; // Number of __m128i vectors to store queryLength short integers.
            size_t nCount = 1 + lenQryShort * alphaSize + (lenQryShort * 6);
            void* pData = SwData::alloc(nCount * sizeof(__m128i ));
            size_t aligned = ((size_t) pData + 15) & ~(0x0f);
            pvsQueryProf = (__m128i *) aligned;
            pvH1 = pvsQueryProf + lenQryShort * alphaSize;
            pvH2 = pvH1 + lenQryShort;
            pvE = pvH2 + lenQryShort;
            pvEGLLoad = pvE + lenQryShort;
            pvEGLStore = pvEGLLoad + lenQryShort;
            pvDIR = pvEGLStore + lenQryShort;
         }
      };

      int swStripedWord(const vindex_t& topSeq, const vindex_t& botSeq, unsigned short gapOpen, unsigned short gapExtend,
                        __m128i *pvQueryProf, __m128i *pvHLoad, __m128i *pvHStore, __m128i *pvE, __m128i *pvEGLLoad, __m128i *pvEGLStore,
                        __m128i *pvDIR, BioSubSequence& top, BioSubSequence& bot) const;

      auto_ptr<SwData> getProfile(const vindex_t& s) const;

      double alignmentImplementation(const vindex_t& topSeq, const vindex_t& botSeq, unsigned short gapOpen, unsigned short gapExtend,
                                     SwData *topProfile, BioSubSequence& top, BioSubSequence& bot) const {
         return swStripedWord(topSeq, botSeq, gapOpen, gapExtend, ((SwFarrarData*) topProfile)->pvsQueryProf,
                              ((SwFarrarData*) topProfile)->pvH1, ((SwFarrarData*) topProfile)->pvH2, ((SwFarrarData*) topProfile)->pvE,
                              ((SwFarrarData*) topProfile)->pvEGLLoad, ((SwFarrarData*) topProfile)->pvEGLStore,
                              ((SwFarrarData*) topProfile)->pvDIR, top, bot);
      }
   public:
      FarrarSWAligner();
      virtual ~FarrarSWAligner();
   };
}
#endif /* FARRARALIGNER_H_ */
