/*
 * MonardoSWAligner.cpp
 *
 *  Created on: Apr 6, 2010
 *      Author: Oleksiy Kohany <kohany@gmail.com>
 */

#include "MonardoSWAligner.h"
#include <memory>
using namespace std;
namespace bio {

   MonardoSWAligner::MonardoSWAligner() {
   }

   MonardoSWAligner::~MonardoSWAligner() {
   }

   auto_ptr<SwData> MonardoSWAligner::getProfile(const vindex_t& s) const {
      auto_ptr<SwMonardoData> pProfile(new SwMonardoData());
      pProfile->alloc(s.size(), scoreMatrix.getAlphabethSize());
      MonardoSWAligner::score_t* wptr = pProfile->queryProfile;
      for (vindex_t::const_iterator sptr = s.begin(); sptr != s.end(); sptr++) {
         for (char c = 0; c < scoreMatrix.getAlphabethSize(); c++) {
            *wptr++ = scoreMatrix.getScore(scoreMatrix.getIndex2Char()[*sptr], scoreMatrix.getIndex2Char()[c]);
         }
      }
      return auto_ptr<SwData> (pProfile);
   }

   double MonardoSWAligner::alignmentImplementation(const vindex_t& topSeq, const vindex_t& botSeq, unsigned short gapOpen,
                                                    unsigned short gapExtend, SwData *topProfile, BioSubSequence& top,
                                                    BioSubSequence& bot) const {
      const size_t topSize = topSeq.size();
      const size_t botSize = botSeq.size();

      const size_t col_size = topSize + 1;
      const size_t row_size = botSize + 1;

      typedef int sumscore_t;
      typedef vector<sumscore_t> vscore_t;
      vscore_t scoreMat(row_size * col_size, 0);
      vtrace_t traceMat(row_size * col_size, -0x8000);

      vector<int> overv(col_size, 0);
      vector<int> upv(row_size, 0);

      sumscore_t maxScore = 0;
      vindex_t::const_iterator ptr1, ptr2;
      score_t* qpRow = ((SwMonardoData*) topProfile)->queryProfile;

      vscore_t::iterator pscore, pscore_nw, pscore_n, pscore_w;
      pscore_nw = scoreMat.begin();
      pscore_n = pscore_nw + 1;
      pscore_w = pscore_nw + row_size;
      pscore = pscore_w + 1;
      vtrace_t::iterator ptrace = traceMat.begin() + row_size;

      int topEnd = 0;
      int botEnd = 0;
      int topIdx, botIdx;

      for (topIdx = 1, ptr1 = topSeq.begin(); ptr1 != topSeq.end(); ++topIdx, ++ptr1, qpRow += scoreMatrix.getAlphabethSize()) {
         ++ptrace;
         for (botIdx = 1, ptr2 = botSeq.begin(); ptr2 != botSeq.end(); ++botIdx, ++ptr2) {
            float bestScore = 0;
            float substScore = *pscore_nw + qpRow[*ptr2];
            if (substScore > 0.0) {
               *pscore = bestScore = substScore;
               *ptrace = 0;
            }
            int k = upv[botIdx];
            int l = topIdx - k;
            float botGapScore = *(scoreMat.begin() + k * row_size + botIdx) - gapCost(gapOpen, gapExtend, l);
            if (botGapScore > bestScore) {
               *pscore = bestScore = botGapScore;
               *ptrace = -l;
            }
            k = overv[topIdx];
            l = botIdx - k;
            float topGapScore = *(scoreMat.begin() + topIdx * row_size + k) - gapCost(gapOpen, gapExtend, l);
            if (topGapScore > bestScore) {
               *pscore = bestScore = topGapScore;
               *ptrace = l;
            }
            if (bestScore > maxScore) {
               maxScore = bestScore;
               topEnd = topIdx;
               botEnd = botIdx;
            }
            if (bestScore == substScore || bestScore == 0.0) {
               if (bestScore >= botGapScore + gapCost(gapOpen, gapExtend, 0)) {
                  upv[botIdx] = topIdx;
               }
               if (bestScore >= topGapScore + gapCost(gapOpen, gapExtend, 0)) {
                  overv[topIdx] = botIdx;
               }
            }
            ++pscore_nw;
            ++pscore_n;
            ++pscore_w;
            ++pscore;
            ++ptrace;
         }
         ++pscore_nw;
         ++pscore_n;
         ++pscore_w;
         ++pscore;
      }
      traceBack(topSeq, botSeq, traceMat, botEnd, topEnd, top, bot);
      return maxScore;
   }
}
