/*
 * MonardoSWAligner.h
 *
 *  Created on: Apr 6, 2010
 *      Author: Oleksiy Kohany <kohany@gmail.com>
 */

#ifndef MONARDOSWALIGNER_H_
#define MONARDOSWALIGNER_H_

#include "SWAligner.h"
#include "math.h"

namespace bio {

   /*
    *
    */
   class MonardoSWAligner: public SWAligner {
      class SwMonardoData: public SwData {
      public:
         SWAligner::score_t *queryProfile;
         SwMonardoData() {
         }
         virtual ~SwMonardoData() {
         }
         void alloc(size_t queryLength, size_t alphaSize) {
            queryProfile = (SWAligner::score_t *) SwData::alloc(queryLength * alphaSize * sizeof(SWAligner::score_t));
         }
      };

      double alignmentImplementation(const vindex_t& topSeq, const vindex_t& botSeq, unsigned short gapOpen, unsigned short gapExtend,
                                     SwData *topProfile, BioSubSequence& top, BioSubSequence& bot) const;
      float gapCost(unsigned short gapOpen, unsigned short gapExtend, int l) const {
         static const int maxGapLength = 500;
         if (l < maxGapLength && l >= 0) {
            static float precalc[maxGapLength];
            if (precalc[l] == 0) {
               precalc[l] = (l == 0 ? gapOpen : gapOpen + gapExtend * log((float) l));
            }
            return precalc[l];
         }
         return l == 0 ? gapOpen : gapOpen + gapExtend * log((float) l);
      }
      auto_ptr<SwData> getProfile(const vindex_t& s) const;
   public:
      MonardoSWAligner();
      virtual ~MonardoSWAligner();
   };
}
#endif /* MONARDOSWALIGNER_H_ */
